package org.polyforms.repository.jpa.executor;

import java.lang.reflect.Method;
import java.util.Arrays;
import java.util.Collections;

import javax.inject.Inject;
import javax.inject.Named;
import javax.persistence.EntityManager;
import javax.persistence.PersistenceContext;
import javax.persistence.Query;

import org.polyforms.repository.jpa.EntityHelper;
import org.polyforms.repository.spi.EntityClassResolver;
import org.polyforms.repository.spi.Executor;

/**
 * Implementation of method which finds entities by identifiers.
 * 
 * @author Kuisong Tong
 * @since 1.0
 */
@Named
public class Find implements Executor {
    private static final String PARAMETER_NAME = "identifiers";
    private static final String SQL_TEMPLATE = "select e from %s e where e.%s in :%s";
    @PersistenceContext
    private EntityManager entityManager;
    private final EntityHelper entityHelper;
    private final EntityClassResolver entityClassResolver;

    /**
     * Create an instance with {@link EntityHelper} and {@link EntityClassResolver}.
     */
    @Inject
    public Find(final EntityHelper entityHelper, final EntityClassResolver entityClassResolver) {
        this.entityHelper = entityHelper;
        this.entityClassResolver = entityClassResolver;
    }

    /**
     * {@inheritDoc}
     */
    public Object execute(final Object target, final Method method, final Object... arguments) {
        Object result = method.isVarArgs() ? Collections.EMPTY_LIST : null;

        if (arguments.length > 0) {
            final Object argument = arguments[0];
            if (method.isVarArgs()) {
                result = findByIds(target, argument);
            } else {
                result = findById(target, argument);
            }
        }

        return result;
    }

    private Object findById(final Object target, final Object argument) {
        return entityManager.find(entityClassResolver.resolve(target.getClass()), argument);
    }

    private Object findByIds(final Object target, final Object argument) {
        final Class<?> entityClass = entityClassResolver.resolve(target.getClass());
        final String queryString = String.format(SQL_TEMPLATE, entityClass.getSimpleName(),
                entityHelper.getIdentifierName(entityClass), PARAMETER_NAME);
        final Query query = entityManager.createQuery(queryString);
        query.setParameter(PARAMETER_NAME, Arrays.asList((Object[]) argument));
        return query.getResultList();
    }
}
